@@ -1,9 +1,11 @@
 /*
- * Copyright © 2009 CNRS, INRIA, Université Bordeaux 1
+ * Copyright © 2009 CNRS
+ * Copyright © 2009-2010 INRIA.  All rights reserved.
+ * Copyright © 2009-2010 Université Bordeaux 1
+ * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
-#include <private/config.h>
 #include <hwloc.h>
 
 #include <stdlib.h>
@@ -14,15 +16,16 @@
 int main(void)
 {
   hwloc_topology_t topology;
-  hwloc_cpuset_t cpuset;
+  hwloc_bitmap_t cpuset;
   hwloc_obj_t obj;
 
   hwloc_topology_init(&topology);
   hwloc_topology_load(topology);
   hwloc_topology_check(topology);
-  cpuset = hwloc_cpuset_alloc();
-  hwloc_cpuset_set(cpuset, 0);
+  cpuset = hwloc_bitmap_alloc();
+  hwloc_bitmap_set(cpuset, 0);
   obj = hwloc_topology_insert_misc_object_by_cpuset(topology, cpuset, "test");
+  hwloc_bitmap_free(cpuset);
   hwloc_topology_insert_misc_object_by_parent(topology, obj, "test2");
   hwloc_topology_check(topology);
   hwloc_topology_destroy(topology);
